@@ -1,6 +1,6 @@
 ;;; mime-edit.el --- Simple MIME Composer for GNU Emacs
 
-;; Copyright (C) 1993,1994,1995,1996,1997,1998 Free Software Foundation, Inc.
+;; Copyright (C) 1993,94,95,96,97,98,99 Free Software Foundation, Inc.
 
 ;; Author: UMEDA Masanobu <umerin@mse.kyutech.ac.jp>
 ;;         MORIOKA Tomohiko <morioka@jaist.ac.jp>
@@ -109,6 +109,7 @@
 (require 'sendmail)
 (require 'mail-utils)
 (require 'mel)
+(require 'eword-encode) ; eword-encode-field-body
 (require 'mime-view)
 (require 'signature)
 (require 'alist)
@@ -120,13 +121,11 @@
 
 (eval-and-compile
   (defconst mime-edit-version
-    (eval-when-compile
-      (concat
-       (mime-product-name mime-user-interface-product) " "
-       (mapconcat #'number-to-string
-		  (mime-product-version mime-user-interface-product) ".")
-       " - \"" (mime-product-code-name mime-user-interface-product) "\"")))
-  )
+    (concat
+     (mime-product-name mime-user-interface-product) " "
+     (mapconcat #'number-to-string
+		(mime-product-version mime-user-interface-product) ".")
+     " - \"" (mime-product-code-name mime-user-interface-product) "\"")))
 
 
 ;;; @ variables
@@ -192,7 +191,7 @@ To insert a signature file automatically, call the function
      ("css") ; rfc2318
      ("xml") ; rfc2376
      ("x-latex")
-     ("x-rot13-47-48")
+     ;; ("x-rot13-47-48")
      )
     ("message"
      ("external-body"
@@ -221,6 +220,7 @@ To insert a signature file automatically, call the function
     ("application"
      ("octet-stream" ("type" "" "tar" "shar"))
      ("postscript")
+     ("vnd.ms-powerpoint")
      ("x-kiss" ("x-cnf")))
     ("image"
      ("gif")
@@ -305,6 +305,11 @@ To insert a signature file automatically, call the function
      "base64"
      "attachment" (("filename" . file))
      )
+    ("\\.ppt$"				; MS Power Point
+     "application" "vnd.ms-powerpoint" nil
+     "base64"
+     "attachment" (("filename" . file))
+     )
 
     ("\\.pln$"
      "text"	"plain"		nil
@@ -485,6 +490,7 @@ If encoding is nil, it is determined from its contents."
     (cn-big5		8 "base64")
     (big5		8 "base64")
     (shift_jis		8 "base64")
+    (tis-620		8 "base64")
     (iso-2022-jp-2	7 "base64")
     (iso-2022-int-1	7 "base64")
     ))
@@ -627,20 +633,27 @@ If it is not specified for a major-mode,
 	  (mime-product-code-name mime-library-product)
 	  ") "
 	  (if (featurep 'xemacs)
-	      (concat (if (featurep 'mule) "MULE")
+	      (concat (cond ((featurep 'utf-2000)
+			     (concat "UTF-2000-MULE/" utf-2000-version))
+			    ((featurep 'mule) "MULE"))
 		      " XEmacs"
-		      (if (string-match "\\s +\\\"" emacs-version)
-			  (concat "/"
-				  (substring emacs-version 0
-					     (match-beginning 0))
-				  (if (and (boundp 'xemacs-betaname)
-					   ;; It does not exist in XEmacs
-					   ;; versions prior to 20.3.
-					   xemacs-betaname)
-				      (concat " " xemacs-betaname)
-				    "")
-				  " (" xemacs-codename ") ("
-				  system-configuration ")")
+		      (if (string-match "^[0-9]+\\(\\.[0-9]+\\)" emacs-version)
+			  (concat
+			   "/"
+			   (substring emacs-version 0 (match-end 0))
+			   (cond ((and (boundp 'xemacs-betaname)
+				       xemacs-betaname)
+				  ;; It does not exist in XEmacs
+				  ;; versions prior to 20.3.
+				  (concat " " xemacs-betaname))
+				 ((and (boundp 'emacs-patch-level)
+				       emacs-patch-level)
+				  ;; It does not exist in FSF Emacs or in
+				  ;; XEmacs versions earlier than 21.1.1.
+				  (format " (patch %d)" emacs-patch-level))
+				 (t ""))
+			   " (" xemacs-codename ") ("
+			   system-configuration ")")
 			" (" emacs-version ")"))
 	    (let ((ver (if (string-match "\\.[0-9]+$" emacs-version)
 			   (substring emacs-version 0 (match-beginning 0))
@@ -675,13 +688,14 @@ inserted into message header.")
 Tspecials means any character that matches with it in header must be quoted.")
 
 (defconst mime-edit-mime-version-value
-  (eval-when-compile
-    (concat "1.0 (generated by " mime-edit-version ")"))
+  (concat "1.0 (generated by " mime-edit-version ")")
   "MIME version number.")
 
 (defconst mime-edit-mime-version-field-for-message/partial
-  (eval-when-compile
-    (concat "MIME-Version: 1.0 (split by " mime-edit-version ")\n"))
+  (concat "MIME-Version:"
+	  (eword-encode-field-body
+	   (concat " 1.0 (split by " mime-edit-version ")\n")
+	   "MIME-Version:"))
   "MIME version field for message/partial.")
 
 
@@ -2163,8 +2177,7 @@ Content-Transfer-Encoding: 7bit
     (goto-char (point-min))
     (while (re-search-forward regexp nil t)
       (delete-region (match-beginning 0)
-		     (progn (forward-line 1) (point)))
-      )))
+		     (1+ (std11-field-end))))))
 
 
 ;;;
@@ -2572,8 +2585,9 @@ Content-Type: message/partial; id=%s; number=%d; total=%d\n%s\n"
 
 (defun mime-edit-quitting-method ()
   "Quitting method for mime-view."
-  (let ((temp mime-raw-buffer)
-	buf)
+  (let* ((entity (get-text-property (point-min) 'mime-view-entity))
+	 (temp (mime-entity-buffer entity))
+	 buf)
     (mime-preview-kill-buffer)
     (set-buffer temp)
     (setq buf mime-edit-buffer)
@@ -2681,7 +2695,7 @@ Content-Type: message/partial; id=%s; number=%d; total=%d\n%s\n"
 	   "^Content-Transfer-Encoding:" limit t)
 	  (let ((beg (match-beginning 0))
 		(hbeg (match-end 0))
-		(end (std11-field-end)))
+		(end (std11-field-end limit)))
 	    (setq encoding
 		  (downcase
 		   (eliminate-top-spaces
@@ -2699,9 +2713,15 @@ Content-Type: message/partial; id=%s; number=%d; total=%d\n%s\n"
 			      encoding nil)
 			)))))))
     (if (or encoded (not not-decode-text))
-	(decode-mime-charset-region (point-min)(point-max)
-				    (or charset default-mime-charset))
-      )
+ 	(progn
+ 	  (save-excursion
+ 	    (goto-char (point-min))
+ 	    (while (re-search-forward "\r\n" nil t)
+ 	      (replace-match "\n")
+ 	      ))
+ 	  (decode-mime-charset-region (point-min)(point-max)
+ 				      (or charset default-mime-charset))
+	  ))
     (let ((he (if (re-search-forward "^$" nil t)
 		  (match-end 0)
 		(point-min)
